from concurrent.futures import ThreadPoolExecutor

from wildcat.annealer.simulated.simulated_annealer import SimulatedAnnealer


class LocalEndpoint:
    def __init__(self, annealer=SimulatedAnnealer()):
        self.annealer = annealer
        pass

    def anneal(self, solver, callback):
        self.annealer.anneal(solver.ising_interactions)
        if not (callback is None):
            callback(self.annealer.q)
        return self.annealer.q

    def dispatch(self, solver, callback=None):
        if solver.ising_interactions.shape[0] == 0:
            raise ValueError('No valid qubo nor ising interactions in the solver.')

        with ThreadPoolExecutor(max_workers=1) as executor:
            return executor.submit(self.anneal, solver, callback)



import numpy as np

from wildcat.annealer.base_annealer import BaseAnnealer
from wildcat.annealer.simulated.single_spin_flip_strategy import SingleSpinFlipStrategy
from wildcat.annealer.simulated.temperature_schedule import TemperatureSchedule
from wildcat.util.matrix import symmetrize


class SimulatedAnnealer(BaseAnnealer):
    def __init__(self, schedule=None, strategy=SingleSpinFlipStrategy()):
        super().__init__()
        self.temperature_schedule = schedule or TemperatureSchedule()
        self.q = np.zeros(0)
        self.strategy = strategy
        self.dim = None
        self.J = None
        self.h = None

    def anneal(self, hamiltonian):
        self.initialize_annealing(hamiltonian)
        for T in self.temperature_schedule:
            self.strategy.update(self, T)
        return self.q

    def initialize_annealing(self, hamiltonian):
        self.dim = hamiltonian.shape[0]
        self.J = symmetrize(hamiltonian).copy()
        self.h = np.zeros(self.dim)
        for i in range(self.dim):
            self.h[i] = hamiltonian[i][i]
            self.J[i][i] = 0
        self.q = np.where(np.random.uniform(-1, 1, self.dim) > 0, 1, -1)



import json
import numpy as np

from requests_futures.sessions import FuturesSession

from wildcat.util.json_encoder import CompactEncoder


class Endpoint:
    def __init__(self, host=None):
        self.host = host or self.default_host
        self.session = FuturesSession()
        self.request_precision = 3

    default_host = "http://api.mdrft.com"
    ising_solver_path = "/apiv1/ising"

    def dispatch(self, solver, path=ising_solver_path, callback=None):
        path = path or self.ising_solver_path
        if not (solver.ising_interactions.shape[0] == 0):
            mat = self._build_matrix_for_params(solver.ising_interactions)
            params = {'hami': np.round(mat, self.request_precision).tolist()}
        else:
            raise ValueError('No valid qubo nor ising interactions in the solver.')

        def handle_result(sess, resp):
            if not (callback is None):
                if resp.status_code != 200:
                    print("Server responded: {}".format(resp.content))
                    callback([])
                else:
                    callback(solver.adjust_solutions_from_ising_spins(np.array(resp.json())))

        request = self.session.post(url=self.host + path, headers={'Content-Type': 'application/json'},
                                    data=json.dumps(params, separators=(',', ':'), cls=CompactEncoder),
                                    background_callback=handle_result)

        return request

    def _build_matrix_for_params(self, matrix, strip=False):
        n = matrix.shape[0]
        list = matrix.tolist()
        if strip:
            for i in range(n):
                list[i] = list[i][i:]
        return list



import numpy as np

from wildcat.annealer.simulated.base_simulated_annealing_strategy import BaseSimulatedAnnealingStrategy


class SingleSpinFlipStrategy(BaseSimulatedAnnealingStrategy):

    def __init__(self, repetition=2000, update_callback=None):
        super().__init__()
        self.repetition = repetition
        self.update_callback = update_callback
        pass

    def update(self, annealer, T):
        for _ in range(self.repetition):
            k = int(np.random.uniform(0, annealer.dim))
            dE = 2 * (annealer.h[k] + annealer.J[k].dot(annealer.q)) * annealer.q[k]
            if dE < 0 or np.exp(- dE / T) > np.random.uniform(0, 1):
                annealer.q[k] *= -1
                if not (self.update_callback is None):
                   self.update_callback(annealer.q)



from wildcat.solver.base_solver import BaseSolver
from wildcat.util.matrix import spin_to_zero_one


class QuboSolver(BaseSolver):
    def __init__(self, qubo=None):
        super().__init__()
        if not (qubo is None):
            self.qubo = qubo
        self.build_ising_interactions()

    def adjust_solutions_from_ising_spins(self, solutions):
        return spin_to_zero_one(solutions)



import numpy as np

from wildcat.network.endpoint import Endpoint
from wildcat.util.matrix import quadratic_energy, hamiltonian_energy


class BaseSolver:
    def __init__(self):
        self.ising_interactions = np.zeros(0)
        self.qubo = np.zeros(0)
        self.endpoint = Endpoint()

    def solve(self, callback, endpoint=None):
        if self.ising_interactions.shape[0] == 0:
            self.qubo()
        if self.qubo.shape[0] == 0:
            self.build_ising_interactions()

        self.endpoint = endpoint or self.endpoint
        return self.endpoint.dispatch(solver=self, callback=callback)

    def build_qubo(self):
        self.qubo = -self.ising_interactions.copy() * 4
        for i in range(self.qubo.shape[0]):
            self.qubo[i][i] = 2 * (self.ising_interactions[i].sum() + self.ising_interactions.sum(axis=0)[i]) - 6 * \
                              self.ising_interactions[i][i]

    def build_ising_interactions(self):
        self.ising_interactions = -self.qubo.copy() / 4.0
        for i in range(self.ising_interactions.shape[0]):
            self.ising_interactions[i][i] = - (self.qubo[i].sum() + self.qubo.sum(axis=0)[i]) / 4.0

    def adjust_solutions_from_ising_spins(self, solutions):
        return solutions

    def hamiltonian_energy(self, spins):
        return hamiltonian_energy(spins, self.ising_interactions)

    def qubo_energy(self, vector):
        return quadratic_energy(vector, self.qubo)



class TemperatureSchedule:
    def __init__(self, scale=0.95, initial_temperature=5.0, last_temperature=0.2):
        self.initial_temperature = initial_temperature
        self.last_temperature = last_temperature
        self.scale = scale
        self.current_temperature = None

    def reset(self):
        self.current_temperature = None

    def __iter__(self):
        if self.current_temperature is None:
            self.current_temperature = self.initial_temperature
        while self.current_temperature > self.last_temperature:
            yield self.current_temperature
            self.current_temperature *= self.scale



import numpy as np


def symmetrize(a):
    return a + a.T - np.diag(a.diagonal())


def check_symmetric(a, tol=1e-8):
    return np.allclose(a, a.T, atol=tol)


def random_symmetric_matrix(size=100):
    return symmetrize(np.random.uniform(-1, 1, (size, size)))


def quadratic_form(vector, matrix):
    return vector.T.dot(matrix).dot(vector)


def zero_one_to_spin(vector):
    return 2 * vector - 1


def spin_to_zero_one(vector):
    return (vector + 1) / 2


def quadratic_energy(vector, matrix):
    product = quadratic_form(vector, matrix)
    for i in range(matrix.shape[0]):
        product += matrix[i][i] * vector[i]
    return product / 2


def hamiltonian_energy(spins, hamiltonian):
    J = hamiltonian.copy()
    for i in range(J.shape[0]):
        J[i][i] = 0
    Jsum = quadratic_form(spins, J) / 2
    hsum = 0
    for i in range(J.shape[0]):
        hsum += hamiltonian[i][i] * spins[i]
    return - (Jsum + hsum)
